--- //depot/vendor/freebsd/src/sys/i386/i386/machdep.c	2007/05/20 22:07:53
+++ //depot/user/attilio/attilio_schedlock/i386/i386/machdep.c	2007/05/26 00:30:54
@@ -1174,7 +1174,8 @@
 	mtx_lock_spin(&dt_lock);
 	if (td->td_proc->p_md.md_ldt)
 		user_ldt_free(td);
-	mtx_unlock_spin(&dt_lock);
+	else
+		mtx_unlock_spin(&dt_lock);
   
 	bzero((char *)regs, sizeof(struct trapframe));
 	regs->tf_eip = entry;
--- //depot/vendor/freebsd/src/sys/i386/i386/sys_machdep.c	2007/05/20 22:07:53
+++ //depot/user/attilio/attilio_schedlock/i386/i386/sys_machdep.c	2007/05/28 21:46:51
@@ -428,7 +428,7 @@
 }
 
 /*
- * Must be called with dt_lock held.
+ * Must be called with dt_lock held.  Returns with dt_lock unheld.
  */
 void
 user_ldt_free(struct thread *td)
@@ -446,6 +446,7 @@
 	}
 
 	mdp->md_ldt = NULL;
+	mtx_unlock_spin(&dt_lock);
 	if (refcount_release(&pldt->ldt_refcnt)) {
 		kmem_free(kernel_map, (vm_offset_t)pldt->ldt_base,
 			pldt->ldt_len * sizeof(union descriptor));
@@ -701,7 +702,7 @@
 		len = NLDT + 1;
 
 	/* Allocate a user ldt. */
-	if ((pldt = mdp->md_ldt) != NULL || len > pldt->ldt_len) {
+	if ((pldt = mdp->md_ldt) == NULL || len > pldt->ldt_len) {
 		struct proc_ldt *new_ldt;
 
 		new_ldt = user_ldt_alloc(mdp, len);
@@ -716,10 +717,13 @@
 				pldt->ldt_sd = new_ldt->ldt_sd;
 				pldt->ldt_base = new_ldt->ldt_base;
 				pldt->ldt_len = new_ldt->ldt_len;
+				mtx_unlock_spin(&dt_lock);
 				kmem_free(kernel_map, (vm_offset_t)old_ldt_base,
 					old_ldt_len * sizeof(union descriptor));
 				FREE(new_ldt, M_SUBPROC);
+				mtx_lock_spin(&dt_lock);
 			} else {
+				mtx_unlock_spin(&dt_lock);
 				/*
 				 * If other threads already did the work,
 				 * do nothing.
@@ -728,14 +732,22 @@
 				   (vm_offset_t)new_ldt->ldt_base,
 				   new_ldt->ldt_len * sizeof(union descriptor));
 				FREE(new_ldt, M_SUBPROC);
+				mtx_lock_spin(&dt_lock);
 				return (0);
 			}
 		} else
 			mdp->md_ldt = pldt = new_ldt;
 #ifdef SMP
-		/* signal other cpus to reload ldt */
+		/*
+		 * Signal other cpus to reload ldt.  We need to unlock dt_lock
+		 * here because other CPU will contest on it since their
+		 * curthreads won't hold the lock and will block when trying
+		 * to acquire it.
+		 */
+		mtx_unlock_spin(&dt_lock);
 		smp_rendezvous(NULL, (void (*)(void *))set_user_ldt_rv,
 		    NULL, td);
+		mtx_lock_spin(&dt_lock);
 #else
 		set_user_ldt(mdp);
 #endif
--- //depot/vendor/freebsd/src/sys/i386/i386/vm_machdep.c	2007/05/20 22:07:53
+++ //depot/user/attilio/attilio_schedlock/i386/i386/vm_machdep.c	2007/05/26 00:30:54
@@ -168,8 +168,8 @@
 				mdp1->md_ldt = pldt;
 				set_user_ldt(mdp1);
 				user_ldt_free(td1);
-			}
-			mtx_unlock_spin(&dt_lock);
+			} else
+				mtx_unlock_spin(&dt_lock);
 		}
 		return;
 	}
@@ -312,8 +312,8 @@
 		td->td_pcb->pcb_gs = _udatasel;
 		load_gs(_udatasel);
 		user_ldt_free(td);
-	}
-	mtx_unlock_spin(&dt_lock);
+	} else
+		mtx_unlock_spin(&dt_lock);
 }
 
 void
--- //depot/vendor/freebsd/src/sys/kern/subr_witness.c	2007/05/20 22:18:03
+++ //depot/user/attilio/attilio_schedlock/kern/subr_witness.c	2007/05/20 23:48:29
@@ -394,6 +394,7 @@
 	{ "sio", &lock_class_mtx_spin },
 #ifdef __i386__
 	{ "cy", &lock_class_mtx_spin },
+	{ "descriptor tables", &lock_class_mtx_spin },
 #endif
 	{ "scc_hwmtx", &lock_class_mtx_spin },
 	{ "uart_hwmtx", &lock_class_mtx_spin },
